//hostmgr.cpp

   /*-----------------------------------------------------------------------*\
   | OWLSock For Windows v1.0                                                |
   --------------------------------------------------------------------------|
   | Written By:  Paul Pedriana                                              |
   | Date:        May 7, 1995.                                               |
   | Copyright:   Copyright (c) 1995 by Paul Pedriana.  All Rights Reserved. |
   | UserID(s):   70541,3223                                                 |
   |              70541.3223@compuserve.com                                  |
   --------------------------------------------------------------------------|
   | OWLSock is a Winsock C++ class library for OWL and Windows (16 and 32). |
   | It makes the use of Winsock much simpler and organizes Winsock's        |
   | functionality into logical classes.  See owlsock.doc and the owlsock    |
   | source files for documentation on how to use OWLSock.                   |
   \*-----------------------------------------------------------------------*/

#include <owl\owlpch.h>
#pragma hdrstop

#include "hostmgr.hpp"
#include "socksub.hpp" //For declaration of INetSocketAddress.


//-********************************************************************
DEFINE_RESPONSE_TABLE1(HostInfoWindow, TWindow)
   EV_MESSAGE(MSG_HOST_INFO_NOTIFY, DoNotification),
END_RESPONSE_TABLE;


//The HostInfoWindow requires a HostInfoManager, so it can relay messages to it.
HostInfoWindow::HostInfoWindow(HostInfoManager* newHostInfoManagerParent)
  :TWindow((TWindow*)NULL, "HostInfo Window"), myHostInfoManagerParent(newHostInfoManagerParent)
{
   Attr.Style = 0; //Turn off WS_CHILD (the default) style).
}

//This is the relay function.
LRESULT HostInfoWindow::DoNotification(WPARAM, LPARAM lParam){
   myHostInfoManagerParent->HostRequestCompleted(WSAGETASYNCERROR(lParam));
   return 1L;
}



//-*********************************************************************************
HostEntry::HostEntry(){
   h_name      = NULL;
   h_aliases   = NULL;
   h_addrtype  = 0;
   h_length    = 0;
   h_addr_list = NULL;
}

int HostEntry::GetAddressCount(){
//Returns the number of pointers to addresses in the hostent (parent class of HostEntry).
   int i=0;

   while(h_addr_list[i])
      i++;
   return i;
}

u_long HostEntry::GetNthINetAddress(int nIndex){
//This function is FLAWED in that it is internet addressing-specific (AF_INET/PF_INET).
//The proper way to implement this function would be to look at h_addrtype and h_length
//  to determine the nature of the address type and return something useful or make a
//  subclass that knows about each address family. (...)
   int nAddressCount = GetAddressCount();

   if(nIndex >= nAddressCount)
      return 0L;  //The caller asked for an address index that is out of range.
   return *((u_long*)h_addr_list[nIndex]);
}


HostInfoManager::HostInfoManager()
   :myWindow(this)
{
   myHostEntry           = (HostEntry*)&chHostInfoBuffer;
   hHostRequest          = 0;
   nLastError            = 0;
   bHostRequestCompleted = 0;
   try{
      myWindow.Create();
   }
   catch(...){
      ::MessageBeep(10);
   }
}



HostInfoManager::~HostInfoManager(){
//In this destructor, we need to clear any pending requests before we get deleted.
   if(hHostRequest && !bHostRequestCompleted){ //If there is an outstanding request...
      CancelHostRequest();
   }
}



int HostInfoManager::GetHostName(char* szName, int nNameLength){
//This function returns the name of the computer on which this program is running. The
//  szName paramter will be set to be the name.
//The return value is either WINSOCK_ERROR or WINSOCK_NOERROR.  You can call the
//  HostInfoManager's GetLastError() to get the actual error value.
//szName is a pointer to pre-allocated buffer of minimum size of nNameLength
//nNameLength should be at least N_MAX_HOST_NAME.
   if(!gethostname(szName, nNameLength)){  //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostAddress(char* szHostAddress, const char* szHostName){
//This function effectively converts szHostName to szHostAddress.  If you have a
//  name such as "joe_schmoe@borland.com", and you want the dotted-decimal IP
//  address for it, you would call this function.
//This function assumes that you have enough space in szHostAddress for the address.
//This function, like most of Winsock 1.1, currently works with only IP addresses.
//  Thus, the szHostAddress is always going to be dotted-decimal format in Winsock 1.1.
//Note that while the inet_ntoa() function has the limitation (feature) that the
//  char* string returned resides in Winsock memory space, the szHostAddress returned
//  is allocated and owned by the caller of this function and can be manipulated any way
//  the caller wants.
//This function returns an error value of WINSOCK_ERROR or WINSOCK_NOERROR.
   int        nError;
   HostEntry* tempHostEntry;

   nError = GetHostInfo(tempHostEntry, szHostName);
   if(nError == WINSOCK_ERROR)
      return nError;
   return HostEntryToAddress(tempHostEntry, szHostAddress);
}



int HostInfoManager::GetHostAddress(SocketAddress& sAddress, const char* szHostName){
//This function effectively converts szHostName to a SocketAddres.  If you have a
//  name such as "joe_schmoe@borland.com", and you want the SocketAddress for it, you
//  could call this function.
//This function, like most of Winsock 1.1, currently works with only IP addresses.
//  Thus, the szHostAddress is always going to be dotted-decimal format in Winsock 1.1.
//This function returns an error value of WINSOCK_ERROR or WINSOCK_NOERROR.
   int        nError;
   HostEntry* tempHostEntry;

   nError = GetHostInfo(tempHostEntry, szHostName);
   if(nError == WINSOCK_ERROR)
      return nError;
   return HostEntryToAddress(tempHostEntry, sAddress);
}



int HostInfoManager::GetHostInfo(HostEntry*& hEntry, const SocketAddress& sAddress){
//The caller of this function is supplying a pointer to be assigned by this function.  Note
//  that the caller need not allocate space for any HostEntry structure.  Note also that
//  because of this, the data needs to be read immediately or copied for later use.
//sAddress is a preallocated SocketAddress reference.  Due to the design of the socket api, the
//  call to gethostbyaddr currently requires a pointer to the internet address, rather than
//  a sockaddr or even a sockaddr_in.  Because passing a u_long pointer would most likely
//  not work if the socket api was to support something other than IP, we fix this issue by
//  making a sockaddr interface to this api.  This comment applies to some other functions
//  in this class as well.
//The address is in network byte ordering.  Note that the call below requires the address in
//  network byte ordering.
   sockaddr_in* tempSockAddrIn;
   tempSockAddrIn = (sockaddr_in*)&sAddress;
   tempSockAddrIn->sin_addr.s_addr = tempSockAddrIn->sin_addr.s_addr;

   hEntry = (HostEntry*)gethostbyaddr((const char*)&tempSockAddrIn->sin_addr.s_addr,
                                         sizeof(u_long), PF_INET);  //sockets call
   if(!hEntry){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostInfo(HostEntry*& hEntry, const char* szName){
//This call is potentially blocking.
//hEntry is a pointer passed by reference.  The system will change that pointer to
//  point to an internal winsock data structure.  You must NOT modify its contents.
//szName is a pre-allocated string that holds a string.  Holds address in text format.
//The caller of this function passes a pointer to a HostEntry struct. e.g.:
//   HostEntry* tempHostEntry;
//   GetHostInfo(tempHostEntry, "JoeShmoe@anywhere.com");
//   printf("%s", tempHostEntry->h_name);  //h_name should be "joeSchmoe@anywhere,com"
   hEntry = (HostEntry*)gethostbyname(szName); //sockets call
   if(!hEntry){
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::CancelHostRequest(HANDLE hTheHostRequest){
//The caller can cancel the last pending request with this call.
   int nTempValue;

   if(!hTheHostRequest)
      hTheHostRequest = hHostRequest;
   if(!hTheHostRequest){
      nLastError=WSAEINVAL;  //There is no handle to use.
      return WINSOCK_ERROR;
   }
   nTempValue = WSACancelAsyncRequest(hTheHostRequest); //sockets call
   if(nTempValue == SOCKET_ERROR){
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostInfoAsync(HANDLE& hTheHostRequest, SocketAddress& sAddress){
//The 'hTheHostRequest' parameter will be returned to the caller with the asynchrous
//  request handle.
//'sAddress' needs to be in network byte ordering.
//Note that due to the design of this class, you cannot have two outstanding Service
//  requests that get notified directly to this class.  You may use the hwnd-specific
//  notification version of this function to manage multiple requests at a time.
//  You may also simply create more than one instance of this class.
//The service is complete when bHostRequestCompleted is TRUE.  You must look at nLastError in
//  this case to see if there was an error. Note also that you must not issue any async
//  calls that post to this class' hwnd until the previous request is completed.
//  The alternative is simply to create multiple HostInfoManagers, or manage the call-backs yourself.
   //For info about the nasty casting done below, see the comments about the non-Async version
   //  of this call.
   sockaddr_in* tempSockAddrIn;
   tempSockAddrIn = (sockaddr_in*)&sAddress;
   tempSockAddrIn->sin_addr.s_addr = tempSockAddrIn->sin_addr.s_addr;

   bHostRequestCompleted=0;
   hHostRequest = hTheHostRequest =
      WSAAsyncGetHostByAddr(myWindow, MSG_HOST_INFO_NOTIFY, (char*)&tempSockAddrIn->sin_addr.s_addr,
                            sizeof(u_long), PF_INET, chHostInfoBuffer, MAXGETHOSTSTRUCT);  //sockets call
   if(!hHostRequest){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostInfoAsync(HANDLE& hTheHostRequest, char* szName){
   bHostRequestCompleted=0;
   hHostRequest = hTheHostRequest =
      WSAAsyncGetHostByName(myWindow, MSG_HOST_INFO_NOTIFY, szName,
                              chHostInfoBuffer, MAXGETHOSTSTRUCT); //sockets call
   if(!hHostRequest){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostInfoAsync(TWindow& wndNotify, HANDLE& hTheHostRequest,
            SocketAddress& sAddress, UINT nMessage, char* chBuffer)
{
//This function notifies the given window about the completion of the request.
//nMessage is the message that the hwndNotify will receive. It defaults to MSG_HOST_INFO_NOTIFY,
//  which is defined in the HostInfoManager's header file.
//The chBuffer is a pointer to buffer that will be filled in with a hostent.  Needs to be at least
//  MAXGETHOSTSTRUCT bytes.  If chBuffer is NULL (or not specified), then the HostInfoManager's
//  internal buffer will be used.
//The hTheHostRequest will hold a handle that the caller can use to reference the request on call-back.
//   wParam will be equal to the hService returned.
//   WSAGETSYNCERROR(lParam)   holds an error, if any (0 is OK).
//   WSAGETSYNCBUFLEN(lParam)  hold actual length of buffer
//When this function returns, myHostEntry will be filled in with the appropriate information.
//  Since this information belongs to this object, you can delay reading it as long as you wish.
//Note that while the sAddress should be passed in network byte ordering, the output on callback
//  will also be in network ordering.
   char* chBufferToUse;
   sockaddr_in* tempSockAddrIn;
   tempSockAddrIn = (sockaddr_in*)&sAddress;
   tempSockAddrIn->sin_addr.s_addr = tempSockAddrIn->sin_addr.s_addr;

   if(chBuffer)
      chBufferToUse = chBuffer;
   else
      chBufferToUse = chHostInfoBuffer;

   bHostRequestCompleted=0;
   hHostRequest = hTheHostRequest =
      WSAAsyncGetHostByAddr(wndNotify, nMessage, (char*)&tempSockAddrIn->sin_addr.s_addr,
                            sizeof(u_long), PF_INET, chBufferToUse, MAXGETHOSTSTRUCT); //sockets call
   if(!hTheHostRequest){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::GetHostInfoAsync(TWindow& wndNotify, HANDLE& hTheHostRequest, char* szName,
                                      UINT nMessage, char* chBuffer)
{
//hwndNotify is the window that will get the message that the request has completed.
//nMessage is the message that the hwndNotify will receive. It defaults to MSG_HOST_INFO_NOTIFY,
//  which is defined in the HostInfoManager's header file.
//hTheHostRequest is the asynchrous request handle that will be a reference to the request.
//szName is the name of the host, as in "coyote@acme.com"
//The chBuffer is a pointer to buffer that will be filled in with a hostent.  Needs to be at least
//  MAXGETHOSTSTRUCT bytes.  If chBuffer is NULL (or not specified), then the HostInfoManager's
//  internal buffer will be used.
//The returned address will be in network byte ordering.
   char* chBufferToUse;

   if(chBuffer)
      chBufferToUse = chBuffer;
   else
      chBufferToUse = chHostInfoBuffer;

   bHostRequestCompleted=0;
   hHostRequest = hTheHostRequest =
      WSAAsyncGetHostByName(wndNotify, nMessage, szName, chBufferToUse, MAXGETHOSTSTRUCT); //sockets call
   if(!hTheHostRequest){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



int HostInfoManager::HostEntryToAddress(HostEntry* hEntry, char* szAddress){
//Given a HostEntry*, this function converts it to a dotted-decimal szAddress. Since
//  Winsock 1.1 supports only IP addressing, this function is limited to IP addressing
//  and the szAddress will always be only dotted-decimal.
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.
   in_addr    tempInAddr;
   char*      szTempAddress;

   tempInAddr.s_addr = *((u_long*)hEntry->h_addr); //else: = *((u_long*)hEntry->h_addr_list[0])
   szTempAddress = inet_ntoa(tempInAddr);          
   if(szTempAddress == NULL)
      return WINSOCK_ERROR;
   strcpy(szAddress, szTempAddress);
   return WINSOCK_NOERROR;
}



int HostInfoManager::HostEntryToAddress(HostEntry* hEntry, SocketAddress& sAddress){
//Given a HostEntry*, this function converts it to a SocketAddress. Since
//  Winsock 1.1 supports only IP addressing, this function is limited to IP addressing
//  and the sAddress will be an INetSocketAddress.
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.
   ((INetSocketAddress)sAddress).SetNetworkAddress(*((u_long*)hEntry->h_addr));
   return WINSOCK_NOERROR;
}



void HostInfoManager::HostRequestCompleted(int nError){
//This function is called whenever an async request is completed.
//You may want to subclass this function.  If you do, you will probably want to call this
//  function as part of the subclassed function.
   if(nError == SOCKET_ERROR)
      nLastError = nError;
   bHostRequestCompleted=1; //Of course, there may have been an error.
}







