//sockets.cpp

   /*-----------------------------------------------------------------------*\
   | OWLSock For Windows v1.0                                                |
   --------------------------------------------------------------------------|
   | Written By:  Paul Pedriana                                              |
   | Date:        May 7, 1995.                                               |
   | Copyright:   Copyright (c) 1995 by Paul Pedriana.  All Rights Reserved. |
   | UserID(s):   70541,3223                                                 |
   |              70541.3223@compuserve.com                                  |
   --------------------------------------------------------------------------|
   | OWLSock is a Winsock C++ class library for OWL and Windows (16 and 32). |
   | It makes the use of Winsock much simpler and organizes Winsock's        |
   | functionality into logical classes.  See owlsock.doc and the owlsock    |
   | source files for documentation on how to use OWLSock.                   |
   \*-----------------------------------------------------------------------*/

#include <owl\owlpch.h>
#pragma hdrstop

#include "sockets.hpp"

//-********************************************************************

//Socket::Socket() is the default constructor for a socket.  You can set the
//  individual members of the Socket later.
Socket::Socket()
   : myWindow(this), nFamily(PF_UNSPEC), nType(0), nProtocol(0),
     nMaxReadBufferSize(N_DEF_MAX_READ_BUFFFER_SIZE), bBound(0),
     bSaveSocket(0), s(INVALID_SOCKET)
{
   ISocket();
}


//Socket(SOCKET&) is a constructor based on a Winsock SOCKET descriptor.
Socket::Socket(SOCKET& newS)
   : myWindow(this), nMaxReadBufferSize(N_DEF_MAX_READ_BUFFFER_SIZE), bBound(0),
     bSaveSocket(0), s(newS)
{
   int nTemp;
   GetMyAddress(mySocketAddress, nTemp, s);   //Fills in our address information.
   nFamily=mySocketAddress.sa_family;
   GetPeerAddress(mySocketAddress, nTemp, s); //Fills in out Peer's address info if it exists.
   ISocket();
}



//This is the standard constructor for a Socket.  It doesn't call socket() nor bind().  You must
//  do those independently.
Socket::Socket(SocketAddress& newSocketAddress, int nNewFamily, int nNewType, int nNewProtocol)
   : myWindow(this), mySocketAddress(newSocketAddress), nFamily(nNewFamily),
     nType(nNewType), nProtocol(nNewProtocol),
     nMaxReadBufferSize(N_DEF_MAX_READ_BUFFFER_SIZE), bBound(0), bSaveSocket(0), s(INVALID_SOCKET)
{
   ISocket();
}



//This Socket destructor will close the socket it if has not be closed already.
//It will also delete the friend notification window as well.
Socket::~Socket(){
  if(s!=INVALID_SOCKET && !bSaveSocket){
     SetLingerOption(0, 60); //Don't wait till data gets sent.  Just kill the socket now.
     CloseSocket();
  }
}



//This function is an intitialization function called by the Socket constructors.  It
//  simply creates the friend window that the Socket needs for Winsock notifications.
void Socket::ISocket(){
//This is for initialization steps that are common to all constructors.
   nLastError = 0;
   try{
      myWindow.Create();
   }
   catch(...){
      //::MessageBox();
   }
}



Socket& Socket::operator=(Socket& newSocket){
//Does a deep copy of the Socket, as much as possible.
   nFamily                 = newSocket.nFamily;
   nType                   = newSocket.nType;
   nProtocol               = newSocket.nProtocol;
   nLastError              = newSocket.nLastError;
   s                       = newSocket.s;
   mySocketAddress         = newSocket.mySocketAddress;
   myPeerSocketAddress     = newSocket.myPeerSocketAddress;
   bSaveSocket             = newSocket.bSaveSocket;

   //Copy the friend (helper) window.  Note that there is an operator= defined
   //  for the class SocketWindow.  However, the SocketWindow::operator= will
   //  set its 'mySocketParent' member to be the newSocket SocketWindow's previous
   //  Socket parent, which may not be us.  Thus, after calling 'myWindow =
   //  newSocket.myWindow;', we call 'myWindow.mySocketParent = this;'.
   myWindow                = newSocket.myWindow;
   myWindow.mySocketParent = this;

   #ifdef B_USE_DATA_QUEUES
   //We copy the queues from the newSocket socket to us here. If the newSocket is in
   //  another (pre-emptively multitasked) thread, and the queues are modified during
   //  this function, then the CopyQueue() calls below could possibly have unusual
   //  results.  Thus, it may be best to either enter a critical section during this
   //  function call or otherwise to make sure the socket will not be sending or
   //  data that could modifythe data queues.  
   CopyQueue(myIncomingQueue, newSocket.myIncomingQueue);
   CopyQueue(myOutgoingQueue, newSocket.myOutgoingQueue);
   #endif //B_USE_DATA_QUEUES
   return *this;
}



int operator==(const Socket& socket1, const Socket& socket2){
//While it is possible that two sockets could refer to the same SOCKET (though
//  this would likely create a mess if not governed with care), it is defined as
//  not possible that two Sockets could lhave the same window member.  THis is
//  because the window is created uniquely on construction for each Socket.
   if(socket1.s != socket2.s)
      return 0;
   if(socket1.myWindow != socket2.myWindow)
      return 0;
   return 1;
}



//See SocketWindow::StartAcceptNotification() for documentation
int Socket::StartAcceptNotification(){
   int nError;

   nError = myWindow.StartAcceptNotification();
   if(nError == WINSOCK_ERROR){
      nLastError = myWindow.GetLastError();
   }
   return nError;
}



//See SocketWindow::StartRegularNotification() for documentation
int Socket::StartRegularNotification(){
   int nError;

   nError = myWindow.StartRegularNotification();
   if(nError == WINSOCK_ERROR){
      nLastError = myWindow.GetLastError();
   }
   return nError;
}


int Socket::StartCustomNotification(int nSelectionOptions){
   int nError;

   nError = myWindow.StartCustomNotification(nSelectionOptions);
   if(nError == WINSOCK_ERROR){
      nLastError = myWindow.GetLastError();
   }
   return nError;
}


//See SocketWindow::CancelNotification() for documentation
int Socket::CancelNotification(){
   int nError;

   nError = myWindow.CancelNotification();
   if(nError == WINSOCK_ERROR){
      nLastError = myWindow.GetLastError();
   }
   return nError;
}



//The SetSocketStyle function can be used to set or change some Socket member data.
//Note that the nNewFamily is also represented in the SocketAddress member, thus they
//  should match.
void Socket::SetSocketStyle(int nNewFamily, int nNewType, int nNewProtocol){
   nFamily=nNewFamily;
   nType=nNewType;
   nProtocol=nNewProtocol;
}



//This function is altogether unreliable.  It makes bogus assumptions about the
//  protocol string.  Try not to use it.  I don't know why I wrote it.
int Socket::ConvertProtocol(char* szNewProtocol){
   if(!szNewProtocol)
      return IPPROTO_TCP;
   if(!strcmp("tcp", szNewProtocol))
      return IPPROTO_TCP;
   if(!strcmp("udp", szNewProtocol))
      return IPPROTO_UDP;
   return IPPROTO_IP;
}



//The CreateSocket function is much like the Winsock socket() function.
//This function assumes that nFamily, nType, and nProtocol are already set properly.
//Note also that since the return of socket() is assigned to 's', that 's' must not
//  be alredy used.  This is another way of saying that there can only be one
//  SOCKET for each Socket object.
int Socket::CreateSocket(){
   s=socket(nFamily, nType, nProtocol);
   if(s==INVALID_SOCKET){
      nLastError=WSAGetLastError();    //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



//The CloseSocket() function is much like the Winsock closesocket() function.
int Socket::CloseSocket(){
   int nErrorValue;

   if(s!=INVALID_SOCKET){
      nErrorValue = closesocket(s);   //sockets call
      if(nErrorValue){
         nLastError = WSAGetLastError();  //sockets call
         return WINSOCK_ERROR;
      }
   }
   s=INVALID_SOCKET; //It's invalid now.
   bBound = 0;
   return WINSOCK_NOERROR;
}



//The ShutDownSocket() function is much like the Winsock shutdown() function.
//It is important to note that shutting down a socket essentially means that
//  you can't un-shut it down.  It is like a graceful way of preparing to
//  end a session.  It is somewhat like a yellow stoplight.
//You would want to use this function to close your socket, while still allowing data
//  be received from the network.  This is as opposed to CloseSocket(), which kills
//  all transfers in both directions.
//nShutMode is one of the enumerations: nShutModeNoRecv, nShutModeNoSend, or nShutModeNoRecvSend.
int Socket::ShutDownSocket(int nShutMode){
   int nErrorValue;

   if(s!=INVALID_SOCKET){
      nErrorValue = shutdown(s, nShutMode);  //sockets call
      if(nErrorValue){
         nLastError = WSAGetLastError();  //sockets call
         return WINSOCK_ERROR;
      }
      s=INVALID_SOCKET; //It's invalid now.
      bBound = 0;
   }
   return WINSOCK_NOERROR;
}




//BindSocket is much like the Winsock bind() function.
//Regardless of what mySocketAddress may have been previously, a call to 'bind()'
//  immediately makes the socket's address the one put into the bind() call.
//  Thus we simply always assign mySocketAddress to be boundSocketAddress.
//The address argument must be in network byte ordering. On the other hand, the
//  SocketAddress class always keeps is addresses in network byte ordering.
int Socket::BindSocket(SocketAddress& addressToBindTo){
   mySocketAddress = addressToBindTo;

   if(bind(s, &mySocketAddress, sizeof(sockaddr))){ //bind() ideally returns 0.
      nLastError=WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   bBound = 1;
   return WINSOCK_NOERROR;
}



//This BindSocket simply binds with our previously defined member data socket address.
int Socket::BindSocket(){
   return BindSocket(mySocketAddress);
}



//This function stores the address into the reference argument 'socketAddress'.
//'nAddressLength' will hold the length of the address.
//'socket' refers to the socket who's address wil be examined.
int Socket::GetMyAddress(SocketAddress& socketAddress, int& nAddressLength, SOCKET& socket){
   if(!getsockname(socket, &socketAddress, &nAddressLength)){
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



//This function stores the address into the reference argument 'socketAddress'.
//'nAddressLength' will hold the length of the address.
//Uses the SOCKET in my member data as the socket to get the address of.
int Socket::GetMyAddress(SocketAddress& socketAddress, int& nAddressLength){
   return GetMyAddress(socketAddress, nAddressLength, s);
}



//The GetPeerAddress() function is much like the Winsock getpeername() function.
//  The Winsock getpeername() function is misnamed; it should be getpeeraddress().
//socketAddress will be changed to have the right addressing info, and nAddressLength
//  will be set to be the address length.
//Note that this function can be used to get the address for any socket descriptor,
//  not just our own socket descriptor.
int Socket::GetPeerAddress(SocketAddress& socketAddress, int& nAddressLength, SOCKET& socket){
   if(socket==0) //This code only works because SOCKET is defined as a u_int and not a struct
      socket=s;

   if(!getpeername(socket, &socketAddress, &nAddressLength)){ //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



//This version of GetPeerAddress() works on our own socket descriptor.
int Socket::GetPeerAddress(SocketAddress& socketAddress, int& nAddressLength){
   return GetPeerAddress(socketAddress, nAddressLength, s);
}



void Socket::SetMyAddress(SocketAddress& newSocketAddress){
//This may be useful for changing the address or setting the address before binding.  It's
//  no good to change this after binding, as a binding is a permanent association between
//  a socket descriptor and a full address (for IP, this is a u_short port and u_long address).  
   mySocketAddress = newSocketAddress;
}


void Socket::SetPeerSocketAddress(SocketAddress& newPeerSocketAddress){
//The 'myPeerSocketAddress' member variable is useful for Datagram sockets because it
//  allows them to specify a default destination to send datagrams to.  With a default
//  destination, a datagram socket that always or often sends to one address can simply call
//  the Write() or Send() functions with no address arguments and the data will send
//  to the default address.
//This function can also be used by a stream socket to set the address for a peer
//  that it wants to connect to.
   myPeerSocketAddress=newPeerSocketAddress;
}




#ifdef B_USE_DATA_QUEUES
//This function does a deep copy of the queue and cleans up the old queue.
void Socket::CopyQueue(DataQueue& queueDestination, DataQueue& queueSource){
   QueueItem* newQueueItem;
   QueueItem* tempQueueItem;
   QueueItemIterator tempQueueItemIterator(myIncomingQueue);
   int i, iLast;

   //Flush the queue and delete each item in it.
   queueDestination.Flush(TShouldDelete::Delete);
   iLast = queueSource.GetItemsInContainer();
   for(i=0; i<iLast; i++){
      newQueueItem = new QueueItem;
      tempQueueItem = tempQueueItemIterator++;
      *newQueueItem = *tempQueueItem;
      queueDestination.Put(newQueueItem);
   }
}
#endif //B_USE_DATA_QUEUES




//GetDriverWaitingSize() is much like calling ioctlsocket(s, FIONREAD,...) in Winsock.
//It returns the number of bytes waiting to be read on our socket.
//For datagrams, it is the size of the next datagram.
//For streams, it "should" be the total waiting bytes.
unsigned long Socket::GetDriverWaitingSize(){
   unsigned long lCharsWaiting;
   int           nErrorValue;

   nErrorValue = ioctlsocket(s, FIONREAD, &lCharsWaiting);  //sockets call
   if(nErrorValue){
      return 0L;
   }
   return lCharsWaiting;
}

#ifdef B_USE_DATA_QUEUES
//This function returns the bytes waiting in the given DataQueue.
unsigned long Socket::GetLocalWaitingSizeQueue(DataQueue& queue){
   QueueItemIterator tempQueueItemIterator(queue);
   QueueItem* tempQueueItem;
   unsigned long lResult;
   int i, iLast;

   lResult=0;
   iLast = myIncomingQueue.GetItemsInContainer();
   for(i=0; i<iLast; i++){
      tempQueueItem = tempQueueItemIterator++;
      lResult += tempQueueItem->nDataLength;
   }
   return lResult;
}
#endif //B_USE_DATA_QUEUES



unsigned long Socket::GetTotalWaitingSize(){
   #ifdef B_USE_DATA_QUEUES
   return GetDriverWaitingSize() + GetLocalWaitingSize();
   #else
   return GetDriverWaitingSize();
   #endif //B_USE_DATA_QUEUES
}



#ifdef B_USE_DATA_QUEUES
unsigned long Socket::GetLocalWaitingSize(){
   return GetLocalWaitingSizeQueue(myIncomingQueue);
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int Socket::GetLocalWaitingCount(){
   return myIncomingQueue.GetItemsInContainer();
}
#endif //B_USE_DATA_QUEUES



Socket::DoReadNotification(const SOCKET& /*socket*/, int /*nError*/){
//This function gets called whenever the socket gets a read notification.  This means
//  that data on the port is ready to be read.
//Thus this function must be subclassed by a DatagramSocket and StreamSocket.
   return 0;
}
Socket::DoWriteNotification(const SOCKET& /*s*/, int /*nError*/){
//The generic socket doesn't know how many bytes it can send, since this limit
//  is dependent on whether the socket is a stream or datagram socket.
//Thus this function must be subclassed by a DatagramSocket and StreamSocket.
   return 0;
}
Socket::DoOOBNotification(const SOCKET& /*s*/, int /*nError*/){
//We don't respond to this in the generic Socket class.
   return 0;
}
Socket::DoAcceptNotification(const SOCKET& /*s*/, int /*nError*/){
//We don't respond to this in the generic Socket class.
   return 0;
}
Socket::DoConnectNotification(const SOCKET& /*s*/, int /*nError*/){
//We don't respond to this in the generic Socket class.
   return 0;
}
Socket::DoCloseNotification(const SOCKET& /*s*/, int /*nError*/){
//We don't respond to this in the generic Socket class.
   return 0;
}

void Socket::SetMaxReadBufferSize(int nNewMaxReadBufferSize){
//This should be called by someone that knows what the correct value is.
   if(nNewMaxReadBufferSize > 0)
      nMaxReadBufferSize = nNewMaxReadBufferSize;
}


#ifdef B_USE_DATA_QUEUES
int Socket::PeekQueue(DataQueue& queue, char* chData, int& nBytesToRead, SocketAddress& sAddress, int nBuffersToRead){
//Reads the bytes into the chData buffer, without removing the data from the queue.
//  Upon exit, nBytesToRead will hold the total bytes that were read. Note that the way the
//  nBytesRead argument works with datagrams is so:
//     If there are more bytes in the incoming datagram than nBytesRead, then the rest of the
//        bytes are lost. If nBytesRead is 0, then no bytes are read into chData, and the
//        queued item is discarded.  This may be useful to rapidly remove a queue item, junk mail.  :)
//     If there are less bytes in the incoming datagram than nBytesRead, then all the bytes
//        that were available are copied.
//sAddress will get set to the address of the sender of the first item in the queue even if
//  more than one buffer was read and the data came from multiple addresses.
//nBuffers to read refers to the total number of queued datagrams to read.  You can put a very high
//  number in here to effectively tell the DatagramSocket to read all buffers up till nBytesToRead.
//  nBuffersToRead defaults to 1, since datagrams are independent communications and can come from
//  different sources.
//Return of 1 means OK, and returns of 0 mean nothing to peek.
   int               nCurrentByteReadPosition, nNewCurrentByteReadPosition, nBytesInCurrentBuffer;
   int               nBytesLeftToRead, nBytesToReadFromCurrentBuffer;
   int               nPresentBuffer, nBuffersInQueue;
   int               i, j, jEnd;
   QueueItem*        tempQueueItem;
   QueueItemIterator tempQueueItemIterator(queue);

   nBuffersInQueue = myIncomingQueue.GetItemsInContainer();
   if(nBuffersToRead > nBuffersInQueue)
      nBuffersToRead = nBuffersInQueue;
   if(!nBuffersToRead)
      return 0;

   sAddress = queue.Peek()->sAddress;
   for(nPresentBuffer=0,nCurrentByteReadPosition=0;
       nPresentBuffer < nBuffersToRead && nCurrentByteReadPosition < nBytesToRead;
       nPresentBuffer++)
   {
      tempQueueItem = tempQueueItemIterator++;
      nBytesInCurrentBuffer = tempQueueItem->nDataLength - tempQueueItem->nPresentPosition;
      nBytesLeftToRead = nBytesToRead - nCurrentByteReadPosition;
      if(nBytesLeftToRead >= nBytesInCurrentBuffer){
         nBytesToReadFromCurrentBuffer = nBytesInCurrentBuffer;  //Read the entire buffer
      }
      else{ //The bytes to read are less than the size of the buffer.
         nBytesToReadFromCurrentBuffer = nBytesLeftToRead;
      }
      nNewCurrentByteReadPosition = nCurrentByteReadPosition + nBytesToReadFromCurrentBuffer;
      for(j=0, i=tempQueueItem->nPresentPosition, jEnd = nBytesToReadFromCurrentBuffer; j<jEnd; j++, i++)
         chData[nCurrentByteReadPosition+j] = tempQueueItem->chData[i];
      nCurrentByteReadPosition = nNewCurrentByteReadPosition;
   }
   nBytesToRead = nCurrentByteReadPosition;
   return 1;
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int Socket::Peek(char* chData, int& nBytesToRead, SocketAddress& sAddress, int nBuffersToRead){
   return PeekQueue(myIncomingQueue, chData, nBytesToRead, sAddress, nBuffersToRead);
}
#endif //B_USE_DATA_QUEUES



Socket::SetBroadcastOption(BOOL bBroadcast){
   if(setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char*)&bBroadcast, sizeof(bBroadcast))){
      nLastError = WSAGetLastError();   //sockets call
      WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetDebugOption(BOOL bDebug){
   if(setsockopt(s, SOL_SOCKET, SO_DEBUG, (char*)&bDebug, sizeof(bDebug))){
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetLingerOption(BOOL bLinger, u_short nLingerTime){
//If you set 'bLinger' to TRUE, then that means to linger for 'nLingerTime' seconds.
//Examples:  bLinger=1, nLingerTime=0.      Hard immediate close. All queued data for sending gets canned immediately.
//           bLinger=1, nLingerTime=2.      Graceful close, waits 2 seconds to try to send any pending data.
//           bLinger=0, nLingerTime=<any>. "Graceful" immediate close, causes data to still be in queue to send when ready.
   LINGER lingerOptions;

   if(bLinger)  //Note that bLinger is a BOOL and LINGER.l_onoff is a u_short.
      lingerOptions.l_onoff = 1;
   else
      lingerOptions.l_onoff = 0;
   lingerOptions.l_linger = nLingerTime;
   if(setsockopt(s, SOL_SOCKET, SO_LINGER, (char*)&lingerOptions, sizeof(lingerOptions))){  //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetRouteOption(BOOL bRoute){
//A False argument means don't route.
   bRoute = !bRoute;
   if(setsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char*)&bRoute, sizeof(bRoute))){ //sockets call
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetKeepAliveOption(BOOL bKeepAlive){
   if(setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char*)&bKeepAlive, sizeof(bKeepAlive))){ //sockets call
      nLastError = WSAGetLastError();     //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetOOBOption(BOOL bSendOOBDataInline){
   if(setsockopt(s, SOL_SOCKET, SO_OOBINLINE, (char*)&bSendOOBDataInline, sizeof(bSendOOBDataInline))){ //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetReceiveBufferOption(int nReceiveBufferSize){
   if(setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&nReceiveBufferSize, sizeof(nReceiveBufferSize))){ //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetReuseAddressOption(int nSendBufferSize){
   if(setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&nSendBufferSize, sizeof(nSendBufferSize))){ //sockets call
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
Socket::SetSendBufferOption(BOOL bAllowReuseAddress){
   if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*)&bAllowReuseAddress, sizeof(bAllowReuseAddress))){ //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}

int Socket::GetBroadcastOption(BOOL& bBroadcast){
   int nSize= sizeof(bBroadcast);
   if(getsockopt(s, SOL_SOCKET, SO_BROADCAST, (char*)&bBroadcast, &nSize)){ //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetDebugOption(BOOL& bDebug){
   int nSize= sizeof(bDebug);
   if(getsockopt(s, SOL_SOCKET, SO_DEBUG, (char*)&bDebug, &nSize)){ //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetLingerOption(BOOL& bLinger, u_short& nLingerTime){
   LINGER lingerOptions;
   int nSize= sizeof(lingerOptions);

   if(getsockopt(s, SOL_SOCKET, SO_LINGER, (char*)&lingerOptions, &nSize)){ //sockets call
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   bLinger = lingerOptions.l_onoff;
   nLingerTime = lingerOptions.l_linger;
   return WINSOCK_NOERROR;
}
int Socket::GetRouteOption(BOOL& bRoute){
   int nSize= sizeof(bRoute);
   if(getsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char*)&bRoute, &nSize)){  //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   //bRoute value of TRUE means "don't route."  So we change it to mean what you'd think.
   bRoute = !bRoute;
   return WINSOCK_NOERROR;
}
int Socket::GetKeepAliveOption(BOOL& bKeepAlive){
   int nSize= sizeof(bKeepAlive);
   if(getsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char*)&bKeepAlive, &nSize)){ //sockets call
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetOOBOption(BOOL& bSendOOBDataInline){
   int nSize= sizeof(bSendOOBDataInline);
   if(getsockopt(s, SOL_SOCKET, SO_OOBINLINE, (char*)&bSendOOBDataInline, &nSize)){ //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetReceiveBufferOption(int& nReceiveBufferSize){
   int nSize= sizeof(nReceiveBufferSize);
   if(getsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&nReceiveBufferSize, &nSize)){ //sockets call
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetSendBufferOption(int& nSendBufferSize){
   int nSize= sizeof(nSendBufferSize);
   if(getsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&nSendBufferSize, &nSize)){ //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}
int Socket::GetReuseAddressOption(BOOL& bAllowReuseAddress){
   int nSize= sizeof(bAllowReuseAddress);
   if(getsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*)&bAllowReuseAddress, &nSize)){ //sockets call
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}



//-********************************************************************
DEFINE_RESPONSE_TABLE1(SocketWindow, TWindow)
   EV_MESSAGE(MSG_SOCKET_NOTIFY, DoNotification),
END_RESPONSE_TABLE;

//Default this to be our standard MSG_SOCKET_NOTIFY
UINT SocketWindow::snMsgSocketNotification = MSG_SOCKET_NOTIFY;

SocketWindow::SocketWindow(Socket* newSocketParent)
  :TWindow((TWindow*)NULL, "Socket Window"), nSelectOptions(0), nNotificationSet(nNotifyAll),
   windowNotification(0), mySocketParent(newSocketParent)
{
   Attr.Style = 0; //Turn off WS_CHILD (the default) style).
}


SocketWindow& SocketWindow::operator =(SocketWindow& newSocketWindow){
//From the user's standpoint, the only thing that is required to make two SocketWindows
//  act as equal is to make their parents and selection options the same.  The actual
//  window handle identities are unimportant.  Thus, we keep our original Window handle,
//  even if the newSocketWindow had an empty window handle.
//However, it may be impossible to assign the parent correctly if this operator is being called
//   in the parent's operator=().  The new parent SHOULD be the original parent, yet the
//   newSocketWindow has a new parent.  We cannot know in this function the conditions
//   under which this assignment is called, so we blindly copy the new parent.
//   The parent will have to override this assignment if the old parent is to
//   remain as it was.
//This function does the best it can to make this window act just like newSocketWindow:
   nSelectOptions     = newSocketWindow.nSelectOptions;
   mySocketParent     = newSocketWindow.mySocketParent;
   windowNotification = newSocketWindow.windowNotification;
   nNotificationSet   = newSocketWindow.nNotificationSet;

   if(mySocketParent->s){
      //Note that if nSelectOptions were empty (0), then we are saying to turn
      //  off notifications and make the socket blocking.
      WSAAsyncSelect(mySocketParent->s, *this, MSG_SOCKET_NOTIFY, nSelectOptions); //sockets call
   }
   return *this;
}

int SocketWindow::StartAcceptNotification(){
//This function says to only listen to FD_ACCEPT messages.  It is important to note that a
//  socket set up to be a listening socket will never be a connected socket, and a connected
//  socket will never receive FD_ACCEPT messages.  Thus all stream sockets are implicitly either
//  connected sockets or listening sockets.
//Since the accepted socket will want to have a different notification window than the
//  listening socket, and the sockets specification says that an accepted socket
//  inherits the notification properties of the listening socket, it is imperative
//  that the listening socket not be set to receive FD_READ, etc notifications.  This
//  is because it is possible that between the accept() call for the new socket and
//  the WSAAsyncSelect() call for the new socket, data may be received for the new socket.
//  Thus the listening socket may get sent the message and it would never get routed to
//  the new socket.
//Calling this function is saying that this SocketWindow is for listening for connections.
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.  You can then examine GetLastError().
   nSelectOptions = FD_ACCEPT;
   return StartCustomNotification(nSelectOptions);
}


int SocketWindow::StartRegularNotification(){
//This function turns on all Winsock notifications except FD_ACCEPT.
//Calling this function is saying that this SocketWindow is for connections rather
//  than for listening.  Since a Winsock socket cannot be a listening socket and
//  a connected socket at the same time, we have serarate the notification functions
//  from each other: StartAcceptNotification() and StartRegularNotification().
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.  You can then examine GetLastError().
   nSelectOptions = FD_ALL & (~FD_ACCEPT);
   return StartCustomNotification(nSelectOptions);
}


int SocketWindow::StartCustomNotification(int nNewSelectOptions){
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.  You can then examine GetLastError().
   int nError;

   nSelectOptions = nNewSelectOptions;
   nError = WSAAsyncSelect(mySocketParent->s, *this, MSG_SOCKET_NOTIFY, nSelectOptions);  //sockets call
   if(nError){
      nLastError = WSAGetLastError();
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}

int SocketWindow::CancelNotification(){
//CancelNotification() turns off the notification to this window. This also changes the
//  socket to be blocking.
//The return value is WINSOCK_ERROR or WINSOCK_NOERROR.  You can then examine GetLastError().
   nSelectOptions = 0;
   return StartCustomNotification(nSelectOptions);
}


//DoNotification() is the SocketWindow's protected internal notification system.
LRESULT SocketWindow::DoNotification(WPARAM wParam, LPARAM lParam){
   SOCKET& theSocket = wParam;
   int     nError    = (int)WSAGETSELECTERROR(lParam);

   if(theSocket != mySocketParent->s)
      return 0;

   switch (WSAGETSELECTEVENT(lParam)){
      case FD_READ:
         if(windowNotification && (nNotificationSet & nNotifyRead)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoReadNotification(theSocket, nError);
         break;
      case FD_WRITE:
         if(windowNotification && (nNotificationSet & nNotifyWrite)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoWriteNotification(theSocket, nError);
         break;
      case FD_OOB:
         if(windowNotification && (nNotificationSet & nNotifyOOB)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoOOBNotification(theSocket, nError);
         break;
      case FD_ACCEPT:
         if(windowNotification && (nNotificationSet & nNotifyAccept)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoAcceptNotification(theSocket, nError);
         break;
      case FD_CONNECT:
         if(windowNotification && (nNotificationSet & nNotifyConnect)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoConnectNotification(theSocket, nError);
         break;
      case FD_CLOSE:
         if(windowNotification && (nNotificationSet & nNotifyClose)){
            return windowNotification->SendMessage(snMsgSocketNotification, wParam, lParam);
         }
         mySocketParent->DoCloseNotification(theSocket, nError);
   }
   return 1;
}



#ifdef B_USE_DATA_QUEUES
//-********************************************************************
//QueueItem
unsigned long QueueItem::lLastSerialNumber = 0;

//Default constructor for a QueueItem. This is a constructor for item
//  that owns nothing, at least not yet.  You can add data to the
//  QueueItem later.
QueueItem::QueueItem()
   :chData(0), nDataLength(0), nPresentPosition(0), bShouldDelete(0),
    lSerialNumber(++lLastSerialNumber)
{
}

//This QueueItem copy constructor does a deep copy of the source constructor.
//Note that while a QueueItem can hold a pointer to data but now own it, a deep
//  copied QueueItem always gets ownership of the data.
QueueItem::QueueItem(QueueItem& newQueueItem){
   CopyQueueItem(newQueueItem);
}


//This is a constructor for newly created QueueItems with no data ready for them
//  yet.  The data presumably will be added later.  Alternatively, you can allocate
//  the data ahead of time, and when it is done and ready, you can create the
//  QueueItem with one of the regular constructors.  Note that in this constructor,
//  the data is allocated, but is simply not yet assigned.
//lLength is the length of the data to hold; bNewShouldDelete is a boolean which
//  identifies whether the QueueItem should take ownership of the data and delete
//  the data in the destructor.
QueueItem::QueueItem(int nLength, short bNewShouldDelete)
   :nDataLength(nLength), bShouldDelete(bNewShouldDelete)
{
   chData = new char[nLength];
}

//Constructor for a QueueItem based on data ready to be held.
//The sNewAddress is the address of the place the data will be sent or where it
//  came from, depending on whether this is an incoming or outgoing QueueItem.
//chNewData is a pointer to the start of the data upon which this QueueItem
//  shoule be based.
//lNewLength is the length of the data to hold, the length of chNewData.
//bNewShouldDelete is a boolean which identifies whether the QueueItem should
//  take ownership of the data and delete the data in the destructor.
//bCopyData means: should the constructor copy the data from chNewData.  Note that
//  it is possible for you to set:
//     bShouldDelete=True  & bCopydata=False
//     bShouldDelete=True  & bCopydata=True
//     bShouldDelete=False & bCopydata=False
//But you should not set:
//     bShouldDelete=False & bCopydata=True
//     (If you copy data, then you must take ownwership of it)
QueueItem::QueueItem(SocketAddress& sNewAddress, char* chNewData, int nNewLength, short bNewShouldDelete, short bCopyData)
   :sAddress(sNewAddress), nDataLength(nNewLength), nPresentPosition(0),
    bShouldDelete(bNewShouldDelete), lSerialNumber(++lLastSerialNumber)
{
   int nCounter;

   if(bCopyData){
      chData = new char[nDataLength];
      for(nCounter=0; nCounter<nDataLength; nCounter++)
         chData[nCounter] = chNewData[nCounter];
   }
   else
      chData = chNewData;  //A shallow copy.
}

//QueueItem destructor.  Simply deletes the data it holds, if it is set to delete it.
QueueItem::~QueueItem(){
   ClearData();
}

//Clears the data the QueueItem data, if it is set to do so.
void QueueItem::ClearData(){
   if(bShouldDelete){
      delete[] chData;
      chData=NULL;
   }
}

//CopyQueueItem() Performs a deep copy of the source item. Note that the
//  bShouldDelete member is set to True, no matter how the sourceQueueItem
//  is set.  This is because we are copying the data and we will therefore
//  be the only object that can own it.
void QueueItem::CopyQueueItem(const QueueItem& sourceQueueItem){
   ClearData(); //Clears the data[] and deallocates it if specified.
   //We don't want to put this: lSerialNumber = sourceQueueItem.lSerialNumber; Intentionally cleared out.
   //  The serial number is unique for ALL QueueItems.
   nDataLength          = sourceQueueItem.nDataLength;
   nPresentPosition     = sourceQueueItem.nPresentPosition;
   chData               = new char[nDataLength];
   if(!chData)
      nDataLength = 0; //Should throw an exception here.
   sAddress      = sourceQueueItem.sAddress;
   bShouldDelete = 1; //I don't think we want put this: bShouldDelete = sourceQueueItem.bShouldDelete;
}

//The operator =() simply copies the QueueItem.  The old data, if it exists,
//  will be deleted.
QueueItem& QueueItem::operator=(const QueueItem& newQueueItem){
   CopyQueueItem(newQueueItem);
   return *this;
}

int operator==(const QueueItem& QueueItem1, const QueueItem& QueueItem2){
//Serial numbers are supposed to be unique, so this function should immediately
//  return with the first 'if' statement right away if unequal.
   if(QueueItem1.lSerialNumber != QueueItem2.lSerialNumber)
      return 0;
   //if(QueueItem1.nDataLength != QueueItem2.nDataLength)
   //   return 0;
   //if(QueueItem1.bShouldDelete != QueueItem2.bShouldDelete)
   //   return 0;
   //if(chData.chData = QueueItem2.chData)
   //   return 0;  //Should I do a deep test of chData?
   return 1;  //They are in fact equal.
}

int operator!=(const QueueItem& QueueItem1, const QueueItem& QueueItem2){
   return !(operator==(QueueItem1, QueueItem2));
}

//The SetAddress() function sets the address that the QueueItem has.  This address
//  is generally the address that an outgoing QueueItem will be sent to, and the
//  address that and incoming QueueItem cam from.
void QueueItem::SetAddress(SocketAddress& sNewAddress){
   sAddress = sNewAddress;
}
#endif //B_USE_DATA_QUEUES



